@@ -49,7 +49,6 @@ void buffer_free(buffer *b) {
 
 __attribute_cold__
 static void buffer_free_ptr(buffer *b) {
-	if (NULL == b) return;
 	free(b->ptr);
 	b->ptr = NULL;
 	b->used = 0;
@@ -57,11 +56,10 @@ static void buffer_free_ptr(buffer *b) {
 }
 
 void buffer_reset(buffer *b) {
-	if (NULL != b && b->size > 0) {
-		b->used = 0;
-		/* release buffer larger than ... bytes */
-		if (b->size > BUFFER_MAX_REUSE_SIZE) buffer_free_ptr(b);
-	}
+	force_assert(NULL != b);
+	b->used = 0;
+	/* release buffer larger than ... bytes */
+	if (b->size > BUFFER_MAX_REUSE_SIZE) buffer_free_ptr(b);
 }
 
 void buffer_move(buffer *b, buffer *src) {
@@ -612,7 +610,6 @@ void buffer_append_string_encoded(buffer *b, const char *s, size_t s_len, buffer
 
 	d = (unsigned char*) buffer_string_prepare_append(b, d_len);
 	buffer_commit(b, d_len); /* fill below */
-	force_assert('\0' == *d);
 
 	for (ds = (unsigned char *)s, d_len = 0, ndx = 0; ndx < s_len; ds++, ndx++) {
 		if (map[*ds]) {
@@ -669,7 +666,6 @@ void buffer_append_string_c_escaped(buffer *b, const char *s, size_t s_len) {
 
 	d = (unsigned char*) buffer_string_prepare_append(b, d_len);
 	buffer_commit(b, d_len); /* fill below */
-	force_assert('\0' == *d);
 
 	for (ds = (unsigned char *)s, d_len = 0, ndx = 0; ndx < s_len; ds++, ndx++) {
 		if ((*ds < 0x20) /* control character */
