@@ -5,6 +5,7 @@
 #include "log.h"
 #include "buffer.h"
 #include "http_header.h"
+#include "stat_cache.h"
 
 #include "plugin.h"
 
@@ -119,7 +120,7 @@ SETDEFAULTS_FUNC(mod_ssi_set_defaults) {
 
 	if (!p) return HANDLER_ERROR;
 
-	p->config_storage = calloc(1, srv->config_context->used * sizeof(plugin_config *));
+	p->config_storage = calloc(srv->config_context->used, sizeof(plugin_config *));
 
 	for (i = 0; i < srv->config_context->used; i++) {
 		data_config const* config = (data_config const*)srv->config_context->data[i];
@@ -584,7 +585,13 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 			}
 		}
 
-		if (0 == stat(p->stat_fn->ptr, &stb)) {
+		if (!con->conf.follow_symlink
+		    && 0 != stat_cache_path_contains_symlink(srv, p->stat_fn)) {
+			break;
+		}
+
+		int fd = stat_cache_open_rdonly_fstat(p->stat_fn, &stb, con->conf.follow_symlink);
+		if (fd > 0) {
 			time_t t = stb.st_mtime;
 
 			switch (ssicmd) {
@@ -619,7 +626,8 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 
 				if (file_path || 0 == p->conf.ssi_recursion_max) {
 					/* don't process if #include file="..." is used */
-					chunkqueue_append_file(con->write_queue, p->stat_fn, 0, stb.st_size);
+					chunkqueue_append_file_fd(con->write_queue, p->stat_fn, fd, 0, stb.st_size);
+					fd = -1;
 				} else {
 					buffer *upsave, *ppsave, *prpsave;
 
@@ -649,6 +657,9 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 
 					con->uri.path = con->physical.rel_path = buffer_init_buffer(srv->tmp_buf);
 
+					close(fd);
+					fd = -1;
+
 					/*(ignore return value; muddle along as best we can if error occurs)*/
 					++p->ssi_recursion_depth;
 					mod_ssi_process_file(srv, con, p, &stb);
@@ -665,6 +676,8 @@ static int process_ssi_stmt(server *srv, connection *con, handler_ctx *p, const
 
 				break;
 			}
+
+			if (fd > 0) close(fd);
 		} else {
 			log_error_write(srv, __FILE__, __LINE__, "sbs",
 					"ssi: stating failed ",
@@ -1168,15 +1181,8 @@ static void mod_ssi_read_fd(server *srv, connection *con, handler_ctx *p, struct
 }
 
 
-/* don't want to block when open()ing a fifo */
-#if defined(O_NONBLOCK)
-# define FIFO_NONBLOCK O_NONBLOCK
-#else
-# define FIFO_NONBLOCK 0
-#endif
-
 static int mod_ssi_process_file(server *srv, connection *con, handler_ctx *p, struct stat *st) {
-	int fd = open(con->physical.path->ptr, O_RDONLY | FIFO_NONBLOCK);
+	int fd = fdevent_open_cloexec(con->physical.path->ptr, con->conf.follow_symlink, O_RDONLY, 0);
 	if (-1 == fd) {
 		log_error_write(srv, __FILE__, __LINE__,  "SsB", "open(): ",
 				strerror(errno), con->physical.path);
