@@ -23,15 +23,6 @@ static void run_buffer_path_simplify(buffer *psrc, buffer *pdest, const char *in
 		fflush(stderr);
 		abort();
 	} else {
-	      #if 0
-		fprintf(stdout,
-			"%s.%d: buffer_path_simplify('%s') succeeded: got '%s'\n",
-			__FILE__,
-			__LINE__,
-			in,
-			out);
-	      #endif
-
 		if (psrc != pdest) buffer_copy_buffer(psrc, pdest);
 		buffer_path_simplify(pdest, psrc);
 
@@ -40,7 +31,7 @@ static void run_buffer_path_simplify(buffer *psrc, buffer *pdest, const char *in
 				"%s.%d: buffer_path_simplify('%s') failed - not idempotent: expected '%s', got '%s'\n",
 				__FILE__,
 				__LINE__,
-				out,
+				in,
 				out,
 				pdest->ptr ? pdest->ptr : "");
 			fflush(stderr);
@@ -51,21 +42,25 @@ static void run_buffer_path_simplify(buffer *psrc, buffer *pdest, const char *in
 
 static void test_buffer_path_simplify_with(buffer *psrc, buffer *pdest) {
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN(""), CONST_STR_LEN(""));
-	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN(" "), CONST_STR_LEN("/"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("/"), CONST_STR_LEN("/"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("//"), CONST_STR_LEN("/"));
-	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc"), CONST_STR_LEN("/abc"));
-	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc//"), CONST_STR_LEN("/abc/"));
-	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc/./xyz"), CONST_STR_LEN("/abc/xyz"));
-	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc/.//xyz"), CONST_STR_LEN("/abc/xyz"));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc"), CONST_STR_LEN("abc"));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc//"), CONST_STR_LEN("abc/"));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc/./xyz"), CONST_STR_LEN("abc/xyz"));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc/.//xyz"), CONST_STR_LEN("abc/xyz"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc/../xyz"), CONST_STR_LEN("/xyz"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("/abc/./xyz"), CONST_STR_LEN("/abc/xyz"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("/abc//./xyz"), CONST_STR_LEN("/abc/xyz"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("/abc/../xyz"), CONST_STR_LEN("/xyz"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc/../xyz/."), CONST_STR_LEN("/xyz/"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("/abc/../xyz/."), CONST_STR_LEN("/xyz/"));
-	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc/./xyz/.."), CONST_STR_LEN("/abc/"));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("abc/./xyz/.."), CONST_STR_LEN("abc/"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("/abc/./xyz/.."), CONST_STR_LEN("/abc/"));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("."), CONST_STR_LEN(""));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN(".."), CONST_STR_LEN(""));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("..."), CONST_STR_LEN("..."));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("...."), CONST_STR_LEN("...."));
+	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN(".../"), CONST_STR_LEN(".../"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("./xyz/.."), CONST_STR_LEN("/"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN(".//xyz/.."), CONST_STR_LEN("/"));
 	run_buffer_path_simplify(psrc, pdest, CONST_STR_LEN("/./xyz/.."), CONST_STR_LEN("/"));
@@ -101,9 +96,62 @@ static void test_buffer_to_lower_upper(void) {
 	buffer_free(psrc);
 }
 
+static void test_buffer_string_space(void) {
+	buffer *b = buffer_init();
+	size_t space;
+
+	space = buffer_string_space(b);
+	assert(0 == space);
+	buffer_copy_string_len(b, CONST_STR_LEN(""));
+	space = buffer_string_space(b);
+	assert(space > 0);
+	assert(space + buffer_string_length(b) == b->size - 1);
+	buffer_commit(b, b->size - 1);
+	assert(b->used == b->size);
+	space = buffer_string_space(b);
+	assert(0 == space);
+
+	buffer_free(b);
+}
+
+static void test_buffer_append_path_len(void) {
+	buffer *b = buffer_init();
+
+	buffer_append_path_len(b, CONST_STR_LEN("a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("/a")));
+	buffer_clear(b);
+	buffer_append_path_len(b, CONST_STR_LEN("a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("/a")));
+	buffer_clear(b);
+	buffer_append_path_len(b, CONST_STR_LEN("/a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("/a")));
+	buffer_copy_string_len(b, CONST_STR_LEN("/"));
+	buffer_append_path_len(b, CONST_STR_LEN("a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("/a")));
+	buffer_copy_string_len(b, CONST_STR_LEN("/"));
+	buffer_append_path_len(b, CONST_STR_LEN("/a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("/a")));
+	buffer_copy_string_len(b, CONST_STR_LEN("a"));
+	buffer_append_path_len(b, CONST_STR_LEN("a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("a/a")));
+	buffer_copy_string_len(b, CONST_STR_LEN("a/"));
+	buffer_append_path_len(b, CONST_STR_LEN("a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("a/a")));
+	buffer_copy_string_len(b, CONST_STR_LEN("a/"));
+	buffer_append_path_len(b, CONST_STR_LEN("/a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("a/a")));
+	buffer_copy_string_len(b, CONST_STR_LEN("/a/"));
+	buffer_append_path_len(b, CONST_STR_LEN("/a"));
+	assert(buffer_is_equal_string(b, CONST_STR_LEN("/a/a")));
+
+	buffer_free(b);
+}
+
 int main() {
 	test_buffer_path_simplify();
 	test_buffer_to_lower_upper();
+	test_buffer_string_space();
+	test_buffer_append_path_len();
 
 	return 0;
 }
