@@ -9,6 +9,7 @@
 #include "request.h"
 #include "base.h"
 #include "burl.h"
+#include "log.h"
 
 static void test_request_connection_reset(connection *con)
 {
@@ -20,9 +21,7 @@ static void test_request_connection_reset(connection *con)
     con->header_len = 0;
     con->http_status = 0;
     buffer_reset(con->proto);
-    buffer_reset(con->parse_request);
     buffer_reset(con->request.request);
-    buffer_reset(con->request.request_line);
     buffer_reset(con->request.orig_uri);
     buffer_reset(con->request.uri);
     array_reset(con->request.headers);
@@ -30,13 +29,14 @@ static void test_request_connection_reset(connection *con)
 
 static void run_http_request_parse(server *srv, connection *con, int line, int status, const char *desc, const char *req, size_t reqlen)
 {
+    int http_status;
     test_request_connection_reset(con);
     buffer_copy_string_len(con->request.request, req, reqlen);
-    http_request_parse(srv, con);
-    if (con->http_status != status) {
+    http_status = http_request_parse(srv, con, con->request.request);
+    if (http_status != status) {
         fprintf(stderr,
                 "%s.%d: %s() failed: expected '%d', got '%d' for test %s\n",
-                __FILE__, line, "http_request_parse", status, con->http_status,
+                __FILE__, line, "http_request_parse", status, http_status,
                 desc);
         fflush(stderr);
         abort();
@@ -260,24 +260,23 @@ static void test_request_http_request_parse(server *srv, connection *con)
       CONST_STR_LEN("GET /\r\n"
                     "\r\n"));
 
-    run_http_request_parse(srv, con, __LINE__, 0,
+    run_http_request_parse(srv, con, __LINE__, 505,
       "zeros in protocol version",
       CONST_STR_LEN("GET / HTTP/01.01\r\n"
                     "Host: foo\r\n"
                     "\r\n"));
-    assert(con->request.http_version == HTTP_VERSION_1_1);
 
-    run_http_request_parse(srv, con, __LINE__, 400,
+    run_http_request_parse(srv, con, __LINE__, 505,
       "missing major version",
       CONST_STR_LEN("GET / HTTP/.01\r\n"
                     "\r\n"));
 
-    run_http_request_parse(srv, con, __LINE__, 400,
+    run_http_request_parse(srv, con, __LINE__, 505,
       "missing minor version",
       CONST_STR_LEN("GET / HTTP/01.\r\n"
                     "\r\n"));
 
-    run_http_request_parse(srv, con, __LINE__, 400,
+    run_http_request_parse(srv, con, __LINE__, 505,
       "strings as version",
       CONST_STR_LEN("GET / HTTP/a.b\r\n"
                     "\r\n"));
@@ -455,16 +454,13 @@ int main (void)
     connection con;
 
     memset(&srv, 0, sizeof(server));
-    srv.errorlog_fd = -1; /* use 2 for STDERR_FILENO from unistd.h */
-    srv.errorlog_mode = ERRORLOG_FD;
-    srv.errorlog_buf = buffer_init();
+    srv.errh = log_error_st_init(&srv.cur_ts, &srv.last_generated_debug_ts);
+    srv.errh->errorlog_fd = -1; /* (disable) */
     srv.split_vals = array_init();
 
     memset(&con, 0, sizeof(connection));
     con.proto                = buffer_init();
-    con.parse_request        = buffer_init();
     con.request.request      = buffer_init();
-    con.request.request_line = buffer_init();
     con.request.orig_uri     = buffer_init();
     con.request.uri          = buffer_init();
     con.request.headers      = array_init();
@@ -476,15 +472,13 @@ int main (void)
     test_request_http_request_parse(&srv, &con);
 
     buffer_free(con.proto);
-    buffer_free(con.parse_request);
     buffer_free(con.request.request);
-    buffer_free(con.request.request_line);
     buffer_free(con.request.orig_uri);
     buffer_free(con.request.uri);
     array_free(con.request.headers);
 
     array_free(srv.split_vals);
-    buffer_free(srv.errorlog_buf);
+    log_error_st_free(srv.errh);
 
     return 0;
 }
